from gtts import gTTS
import os


class Narrator:
    def __init__(self, class_labels: list, audio_directory: str, max_obj_per_segment=5,
                 rank_labels=('close', 'near', 'far'), h_direction_labels=('left', 'middle', 'right'),
                 v_direction_labels=('above', 'midst', 'bottom'), horizontal_only=True):
        """
        :param class_labels: list of class labels
        :param audio_directory: path of desired directory for audio narrations to be downloaded. if not exists, it will be created
        :param max_obj_per_segment: the maximum number of objects in a single narration
        :param rank_labels: labels associated with the ranks of the objects. Tuple of length more than zero and less than five to cover the possible quartiles in ranks
        :param h_direction_labels: labels associated with the horizontal location of the objects. Tuple of length three
        :param v_direction_labels: labels associated with the vertical location of the objects. Tuple of length three
        :param horizontal_only: whether to download annotations regarding only the horizontal location of the objects or not
        """
        if not os.path.isdir(audio_directory):
            os.mkdir(audio_directory)
        self._audio_directory = audio_directory
        self._horizontal_only = horizontal_only
        if not audio_directory.endswith(('/', '\\')):
            audio_directory += '/'
        for c in class_labels:
            if horizontal_only:
                for distance in rank_labels:
                    for direction in h_direction_labels:
                        text = f'one {c} {direction}-{distance}'
                        tts = gTTS(text, lang='en')
                        tts.save(f'{audio_directory}1-{c}-{direction}-{distance}.mp3')
            else:
                for distance in rank_labels:
                    for v_direction in v_direction_labels:
                        for h_direction in h_direction_labels:
                            text = f'one {c} {h_direction} {v_direction} {distance}'
                            tts = gTTS(text, lang='en')
                            tts.save(f'{audio_directory}1-{c}-{h_direction}-{v_direction}-{distance}.mp3')
            for i in range(2, max_obj_per_segment + 1):
                if horizontal_only:
                    for distance in rank_labels:
                        for direction in h_direction_labels:
                            text = f'{i} {c}s  {direction} {distance}'
                            tts = gTTS(text, lang='en')
                            tts.save(f'{audio_directory}{i}-{c}-{direction}-{distance}.mp3')
                else:
                    for distance in rank_labels:
                        for v_direction in v_direction_labels:
                            for h_direction in h_direction_labels:
                                text = f'{i} {c}s  {h_direction} {v_direction} {distance}'
                                tts = gTTS(text, lang='en')
                                tts.save(f'{audio_directory}{i}-{c}-{h_direction}-{v_direction}-{distance}.mp3')

    @staticmethod
    def get_narration(classes, class_labels, ranks, locations, rank_to_distance_labels=('far', 'near', 'near', 'close'),
                      h_location_to_lr_labels=('left', 'middle', 'right'),
                      v_location_to_ab_labels=('above', 'midst', 'bottom'),
                      horizontal_only=True):
        """
        :param classes: array of shape (n, ) with each element corresponding to the index of the object's class in class_labels
        :param class_labels: list of class labels of objects in your dataset. For example: ['car', 'bike', 'person', 'truck'].
ranks: output of tv.Ruler.get_rank(). Array of shape (n, ) with each element being 0, 1, 2, or 3 representing an object's size/distance
        :param ranks: output of tv.locate(). Array of shape (n, 2) containing the locations of the boxes: [h_location, v_location]
        :param locations: output of tv.locate(). Array of shape (n, 2) containing the locations of the boxes: [h_location, v_location]
        :param rank_to_distance_labels: tuple of length 4 describing objects' distance/size according to their ranks (quartile intervals)
        :param h_location_to_lr_labels: tuple of length 3 describing objects' horizontal location
        :param v_location_to_ab_labels: tuple of length 3 describing objects' vertical location
        :param horizontal_only: whether to create narrations regarding objects' vertical location. Must be set to True if a narrator's audio files were created with it set true in tv.Narrator()
        :return: a list of narrations that correspond to the downloaded audio files for a narrator which can also be used in a textual format
        """
        map_ = {}
        for i, c in enumerate(classes):
            label = class_labels[c]
            key = f'-{label}-{h_location_to_lr_labels[locations[i][0]]}'
            if not horizontal_only:
                key += f'-{v_location_to_ab_labels[locations[i][1]]}'
            key += f'-{rank_to_distance_labels[ranks[i]]}'
            if key in map_:
                map_[key] += 1
            else:
                map_[key] = 1
        narration = []
        for key in map_:
            narration.append(f'{map_[key]}{key}')
        return narration



import numpy as np


class Ruler:
    def __init__(self, images: list, class_labels: list, coords_key='box', class_key='class', xmin_key='x1', ymin_key='y1',
                 xmax_key='x2', ymax_key='y2'):
        """
        fit the ruler on a collection of bounding box annotations for each class to get quantiles of the area of objects
        :param images: list of dicts where each one corresponds to bounding box annotations of an image i.e. [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param class_labels: list of class labels
        :param coords_key: alternative key for 'box' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param class_key: alternative key for 'class' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param xmin_key: alternative key for 'x1' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param ymin_key: alternative key for 'y1' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param xmax_key: alternative key for 'x2' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        :param ymax_key: alternative key for 'y2' in [{'box':{'x1': 0, 'y1': 10, 'x2': 10, 'y2':20}, 'class': 'car'}]
        """
        self._quantiles = {c: [] for c in class_labels}
        self._class_labels = class_labels
        for image in images:
            for obj in image:
                coords = obj[coords_key]
                class_ = obj[class_key]
                w = coords[xmin_key] - coords[xmax_key]
                h = coords[ymin_key] - coords[ymax_key]
                self._quantiles[class_].append(w * h)
        for c in self._quantiles:
            self._quantiles[c] = np.quantile(self._quantiles[c], [.25, .5, .75])

    def get_ranks(self, boxes, classes):
        """
        :param boxes: array-like object of shape (n_boxes, 4) with each element containing [ymin, xmin, ymax, xmax] coordinates of bounding boxes
        :param classes: array-like object of shape (n_boxes, ) with each element corresponding to the index of the object's class in class labels
        :return: an array of objects' ranks where a rank corresponds to the quantile interval of the object's size among the objects fitted in initialization -> 0, 1, 2, or 3
        """
        ranks = []
        for box, class_index in zip(boxes, classes):
            ymin, xmin, ymax, xmax = box
            class_ = self._class_labels[class_index]
            size = (ymax - ymin) * (xmax - xmin)
            ranks.append(np.searchsorted(self._quantiles[class_], size))
        return ranks



import numpy as np


def locate(boxes, scene_width, scene_height, v_point=.66, h_point=.66, horizontal_only=True):
    """
    :param boxes: numpy array of shape (n_boxes, 4) with each element containing [ymin, xmin, ymax, xmax] coordinates of bounding boxes
    :param scene_width: the pixel-width of the scene to be analyzed
    :param scene_height: the pixel-height of the scene to be analyzed
    :param v_point: the portion of a box's width, as a threshold to determine its vertical location
    :param h_point: the portion of a box's height, as a threshold to determine its horizontal location
    :param horizontal_only: defaults to True, if False, horizontal location of the boxes will be determined as well
    :return: an array of [horizontal, vertical] location of each box, 0 for left (above), 1 for middle (midst), 2 for right (bottom), and -1 in vertical index if horizontal_only=True
    """
    where = np.ones((boxes.shape[0], 2), dtype='int8')

    v_margin = scene_width / 2
    both_left = (boxes[:, 1] <= v_margin) & (boxes[:, 3] <= v_margin)
    both_right = np.full(both_left.shape, False)
    both_right[~both_left] = (boxes[~both_left, 1] >= v_margin) & (boxes[~both_left, 3] >= v_margin)
    v_threshold = (boxes[:, 3] - boxes[:, 1]) * h_point
    left_h_margin = v_margin - v_threshold
    left = np.full(both_left.shape, False)
    left[~both_left & ~both_right] = boxes[~both_left & ~both_right, 1] < left_h_margin[~both_left & ~both_right]
    right_h_margin = v_margin + v_threshold
    right = np.full(both_left.shape, False)
    right[~both_left & ~both_right & ~left] = boxes[~both_left & ~both_right & ~left, 3] > right_h_margin[
        ~both_left & ~both_right & ~left]
    where[np.where(both_left | left), 0] = 0
    where[np.where(both_right | right), 0] = 2

    if not horizontal_only:
        v_margin = scene_height / 2
        both_above = (boxes[:, 0] <= v_margin) & (boxes[:, 2] <= v_margin)
        both_bottom = np.full(both_above.shape, False)
        both_bottom[~both_above] = (boxes[~both_above, 0] >= v_margin) & (boxes[~both_above, 2] >= v_margin)
        v_threshold = (boxes[:, 2] - boxes[:, 0]) * v_point
        bottom_v_margin = v_margin - v_threshold
        above = np.full(both_above.shape, False)
        above[~both_above & ~both_bottom] = boxes[~both_above & ~both_bottom, 0] < bottom_v_margin[~both_above & ~both_bottom]
        above_v_margin = v_margin + v_threshold
        bottom = np.full(both_bottom.shape, False)
        bottom[~both_above & ~both_bottom & ~bottom] = boxes[~both_above & ~both_bottom & ~bottom, 2] > above_v_margin[
            ~both_above & ~both_bottom & ~bottom]
        where[np.where(both_above | above), 1] = 0
        where[np.where(both_bottom | bottom), 1] = 2
    else:
        where[..., 1] = -1

    return where



from setuptools import setup

setup(
    name='tell_a_vision',
    version='0.1.7',
    description='TV (tell_a_vision) is an inference engine for object detection that provides explanatory analysis for such tasks in computer vision.',
    url='https://github.com/rezmansouri/tell_a_vision',
    author='Reza Mansouri',
    author_email='std_reza_mansouri@khu.ac.ir',
    license='MIT',
    install_requires=['numpy >= 1.21.0',
                      'gTTs >= 2.2.4'
                      ],
    long_description="TV (tell_a_vision) is an inference that takes object detection algorithms' (such as YOLO) output as input and provides explanatory data such as the location, number, and size (distance) of the objects found with respect to their classification in a descriptive format along with audio output. You can try it using this colab notebook demo: https://colab.research.google.com/drive/1o6MgntmIb1qLLpGXxsDw0W6SMH2kheCZ?usp=sharing."
)
